use bootloader::bootinfo::{MemoryMap, MemoryRegionType};
use linked_list_allocator::{Heap, LockedHeap};
use spinning_top::{lock_api::MutexGuard, RawSpinlock};
use x86_64::{
	registers::control::Cr3,
	structures::paging::{
		mapper::MapToError, page_table::FrameError, FrameAllocator, Mapper, OffsetPageTable, Page, PageTable,
		PageTableFlags, PhysFrame, Size4KiB,
	},
	PhysAddr, VirtAddr,
};

use crate::println;

pub fn init(physical_memory_start: u64, mapping: &'static MemoryMap) {
	let mut mapper = unsafe { init_mapper(physical_memory_start) };
	let mut frame_allocator = unsafe { BootInfoFrameAllocator::new(mapping) };
	init_heap(&mut mapper, &mut frame_allocator).unwrap();
}

pub unsafe fn init_mapper(physical_memory_start: u64) -> OffsetPageTable<'static> {
	println!("Initializing Mapper.");
	let physical_memory_start = VirtAddr::new(physical_memory_start);
	let level_4_table = active_level_4_table(physical_memory_start);
	OffsetPageTable::new(level_4_table, physical_memory_start)
}

unsafe fn active_level_4_table(physical_memory_start: VirtAddr) -> &'static mut PageTable {
	let (l4_table, _flags) = Cr3::read();
	let l4_start = l4_table.start_address();
	let virt = physical_memory_start + l4_start.as_u64();
	let page_table_ptr = virt.as_mut_ptr::<PageTable>();
	&mut *page_table_ptr
}

pub struct BootInfoFrameAllocator {
	memory_map: &'static MemoryMap,
	next: usize,
}

impl BootInfoFrameAllocator {
	pub unsafe fn new(memory_map: &'static MemoryMap) -> Self {
		println!("Initializing Frame Allocator.");
		BootInfoFrameAllocator { memory_map, next: 0 }
	}

	fn usable_frames(&self) -> impl Iterator<Item = PhysFrame> {
		let regions = self.memory_map.iter();
		let usable_regions = regions.filter(|r| r.region_type == MemoryRegionType::Usable);
		let addr_ranges = usable_regions.map(|r| r.range.start_addr()..r.range.end_addr());
		let frame_addresses = addr_ranges.flat_map(|r| r.step_by(4096));
		frame_addresses.map(|addr| PhysFrame::containing_address(PhysAddr::new(addr)))
	}
}

unsafe impl FrameAllocator<Size4KiB> for BootInfoFrameAllocator {
	fn allocate_frame(&mut self) -> Option<PhysFrame> {
		let frame = self.usable_frames().nth(self.next);
		self.next += 1;
		frame
	}
}

pub const HEAP_START: usize = 0x_4444_4444_0000;
pub const HEAP_SIZE: usize = 10_000 * 1024; // 10 MiB

pub fn init_heap(
	mapper: &mut impl Mapper<Size4KiB>,
	frame_allocator: &mut impl FrameAllocator<Size4KiB>,
) -> Result<(), MapToError<Size4KiB>> {
	println!("Initializing Heap.");
	let page_range = {
		let heap_start = VirtAddr::new(HEAP_START as u64);
		let heap_end = heap_start + HEAP_SIZE as u64 - 1u64;
		let heap_start_page = Page::containing_address(heap_start);
		let heap_end_page = Page::containing_address(heap_end);
		Page::range_inclusive(heap_start_page, heap_end_page)
	};

	for page in page_range {
		let frame = frame_allocator
			.allocate_frame()
			.ok_or(MapToError::FrameAllocationFailed)?;
		let flags = PageTableFlags::PRESENT | PageTableFlags::WRITABLE;
		unsafe { mapper.map_to(page, frame, flags, frame_allocator)?.flush() };
	}

	unsafe { alloc().init(HEAP_START as _, HEAP_SIZE) };

	Ok(())
}

#[global_allocator]
static ALLOCATOR: LockedHeap = LockedHeap::empty();

pub fn alloc<'l>() -> MutexGuard<'l, RawSpinlock, Heap> {
	ALLOCATOR.lock()
}

pub unsafe fn translate_addr(addr: VirtAddr, physical_memory_start: VirtAddr) -> Option<PhysAddr> {
	translate_addr_inner(addr, physical_memory_start)
}

fn translate_addr_inner(addr: VirtAddr, physical_memory_start: VirtAddr) -> Option<PhysAddr> {
	let (level_4_table_frame, _) = Cr3::read();
	let table_indexes = [addr.p4_index(), addr.p3_index(), addr.p2_index(), addr.p1_index()];
	let mut frame = level_4_table_frame;
	for &index in &table_indexes {
		let virt = physical_memory_start + frame.start_address().as_u64();
		let table_ptr: *const PageTable = virt.as_ptr();
		let table = unsafe { &*table_ptr };
		let entry = &table[index];
		frame = match entry.frame() {
			Ok(frame) => frame,
			Err(FrameError::FrameNotPresent) => return None,
			Err(FrameError::HugeFrame) => panic!("huge pages not supported."),
		};
	}
	Some(frame.start_address() + u64::from(addr.page_offset()))
}
