use core::mem;

use alloc::{
	string::{String, ToString},
	vec::Vec,
};
use blinkcast::alloc::channel;
use microlang::eval::{Context, Scope, Value};

use crate::{
	dev::{console::out, memory::alloc},
	hard::keyboard::{keyboard, Code, Key, KeyModState},
	print, println,
};

use crate::hard::keyboard::Azerty as Keymap;
// use crate::hard::keyboard::Qwerty as Keymap;

pub fn main() {
	let keymap = Keymap::default();

	let (tx, mut rx) = channel(1024);
	let mut buff = String::new();
	let mut state = KeyModState::default();
	keyboard().keymap(keymap.clone());
	keyboard().on_code(move |c| match state.translate(&keymap, c) {
		Code::Down(Key::Char('d')) if state.ctrl => {
			panic!("Exiting.");
		}
		Code::Down(Key::Char(c)) => {
			print!("{c}");
			if c == '\n' {
				let mut new = String::new();
				mem::swap(&mut new, &mut buff);
				tx.send(new);
			} else {
				buff.push(c);
			}
		}
		Code::Down(Key::BackSpace) => {
			out().correct();
			buff.pop();
		}
		_ => (),
	});

	let mut context = Context::empty();
	microlang::prelude::prelude(&mut context);
	prelude(&mut context);

	print!("> ");
	loop {
		if let Some(line) = rx.recv() {
			match context.eval(line) {
				Ok((_, value)) => println!("{}", value.serialize()),
				Err(err) => println!("Error: {err:?}"),
			}
			print!("> ");
		}
	}
}

fn prelude(context: &mut Context) {
	context.define_built("out", |args, _ctx| {
		let mut first = true;
		for arg in args {
			let str = arg.serialize();
			if first {
				print!("{str}");
				first = false;
			} else {
				print!(" {str}");
			}
		}
		println!();
		Ok(Value::None)
	});

	context.define_built("env", |_args, ctx| {
		fn rec(scope: &Scope) -> Vec<(String, String)> {
			scope
				.defs
				.iter()
				.map(|(key, value)| (key.to_string(), value.serialize()))
				.chain(
					scope
						.parent
						.clone()
						.map(|p| rec(&p.borrow()).into_iter())
						.into_iter()
						.flatten(),
				)
				.collect()
		}
		let values = rec(&ctx.frame.borrow())
			.into_iter()
			.map(|(n, v)| Value::array([Value::str(n), Value::str(v)].to_vec()))
			.collect();
		Ok(Value::array(values))
	});

	context.define_built("mem", |_args, _ctx| {
		let size = alloc().size() as f64;
		let used = alloc().used() as f64;
		let free_frac = 1. - (used / size);
		Ok(Value::object(
			[
				("size".into(), Value::Num(size)),
				("used".into(), Value::Num(used)),
				("free_frac".into(), Value::Num(free_frac)),
			]
			.into_iter(),
		))
	});
}
